"""
This package contains helpers for dealing with digest headers
in the various Fediverse settings.

See [RFC 9530 Digest Fields](https://www.rfc-editor.org/rfc/rfc9530.html)
for the current relevant RFC.
"""

import http_sf
import hashlib
import logging

from .helper import content_digest_sha256
from .multibase import multibase_58btc_encode

logger = logging.getLogger(__name__)


def validate_digest(headers: dict, body: bytes) -> bool:
    """Validates the digest. First checks the `digest` header
    then the `content-digest` header.

    ```pycon
    >>> validate_digest({"digest": "sha-256=Kch/yJ/aOjLud24QANj5EK/SfmpAubIsE9BbRcaT5D4="},
    ...     b'mooo')
    True

    >>> validate_digest({"digest": "SHA-256=Kch/yJ/aOjLud24QANj5EK/SfmpAubIsE9BbRcaT5D4="},
    ...     b'mooo')
    True

    >>> validate_digest({"content-digest": 'sha-256=:R9+ukoir89XSJSq/sL1qyWYmN9ZG5t+dXSdLwzbierw=:'},
    ...     b'moo')
    True

    ```

    :param headers: The headers of the request
    :param body: The body of the request, currently a warning is raised if body is of type str
    :return: True if digest is present and valid
    """

    if "digest" in headers:
        request_digest = headers["digest"]
        request_digest = request_digest[:4].lower() + request_digest[4:]
        digest = content_digest_sha256(body)
        if request_digest != digest:
            logger.warning("Different digest")
            return False

        return True

    if "content-digest" in headers:
        try:
            parsed = http_sf.parse(
                headers["content-digest"].encode("utf-8"), tltype="dict"
            )
        except Exception as e:
            logger.warning(
                "Failed to parse header %s with %s",
                headers["content-digest"],
                repr(e),
            )
            return False

        if len(set(parsed.keys()) - {"sha-256", "sha-512"}) > 0:
            logger.warning(
                "Got unsupported hash method in %s", headers["content-digest"]
            )

        valid = False

        if "sha-256" in parsed:
            if parsed["sha-256"][0] == hashlib.sha256(body).digest():
                valid = True
            else:
                return False

        if "sha-512" in parsed:
            if parsed["sha-512"][0] == hashlib.sha512(body).digest():
                valid = True
            else:
                return False

        return valid

    return False


def digest_multibase(obj: bytes) -> str:
    """
    Implements the multibase multihash digest, see [here](https://github.com/multiformats/multihash).
    This was proposed to use in the Fediverse in
    [FEP-ef61: Portable Objects](https://codeberg.org/fediverse/fep/src/branch/main/fep/ef61/fep-ef61.md).

    ```pycon
    >>> digest_multibase(b"multihash")
    'zQmYtUc4iTCbbfVSDNKvtQqrfyezPPnFvE33wFmutw9PBBk'

    ```
    """

    return multibase_58btc_encode(b"\x12\x20" + hashlib.sha256(obj).digest())



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from urllib.parse import urlparse

uris_for_public = [
    "Public",
    "as:Public",
    "https://www.w3.org/ns/activitystreams#Public",
]
"""URIs used to represent the public collection. See [Section 5.6 of ActivityPub](https://www.w3.org/TR/activitypub/#public-addressing)."""


def actor_for_object(data: dict) -> str:
    """Look up for the actor id either from attributedTo or actor"""
    if "attributedTo" in data:
        actor = data.get("attributedTo")
    else:
        actor = data.get("actor")
    actor = id_for_object(actor)

    if actor:
        return actor

    return "__NO__ACTOR__"


def id_for_object(data: dict | str | None) -> str | None:
    """Determines the id of an object

    ```pycon
    >>> id_for_object("http://obj.example")
    'http://obj.example'

    >>> id_for_object({"id": "http://obj.example"})
    'http://obj.example'

    >>> id_for_object({"name": "alice"}) is None
    True

    ```
    """

    if data is None:
        return None
    if isinstance(data, str):
        return data
    return data.get("id", None)


def property_for_key_as_set(data, key):
    """Returns value as a set, useful for `to` and `cc`

    ```pycon
    >>> property_for_key_as_set({"to": "http://actor.example"}, "to")
    {'http://actor.example'}

    ```
    """
    if data is None:
        return set()
    result = data.get(key, [])
    if isinstance(result, str):
        return set([result])
    return set(result)


def recipients_for_object(data: dict) -> set:
    """Combines the recipients from to, cc, bto, bcc, audience into a set

    ```pycon
    >>> result = recipients_for_object({
    ...     "to": ["http://to.example"],
    ...     "cc": "http://cc.example",
    ...     "bcc": ["http://bcc.example"],
    ...     "audience": "http://audience.example"})
    >>> sorted(result)
    ['http://audience.example', 'http://bcc.example',
        'http://cc.example', 'http://to.example']


    ```

    !!! note
        treatment of audience might change.

    """
    return set.union(
        *[
            property_for_key_as_set(data, key)
            for key in ["to", "cc", "bto", "bcc", "audience"]
        ]
    )


def remove_public(recipients):
    """Given a list of Recipients removes public

    ```pycon
    >>> remove_public(["Public", "as:Public",
    ...     "https://www.w3.org/ns/activitystreams#Public",
    ...     "http://alice.example"])
    {'http://alice.example'}

    ```
    """

    return {x for x in recipients if x not in uris_for_public}


def contains_public(array: list[str]) -> bool:
    """Checks if the list contains public

    ```pycon
    >>> contains_public(["Public"])
    True

    >>> contains_public(["http://alice.example"])
    False

    ```
    """

    return any(x in uris_for_public for x in array)


def is_public(data: dict) -> bool:
    """Determines if the object should be considered public based on its recipients"""
    return contains_public(recipients_for_object(data))


def fediverse_handle_from_actor(actor: dict) -> str:
    """Given an actor object, i.e. a dict, determines the fediverse handle"""
    host = urlparse(actor["id"]).netloc
    username = urlparse(actor["id"]).path.split("/")[-1]

    if "preferredUsername" in actor:
        username = actor["preferredUsername"]

    return f"{username}@{host}"


def copy_to_and_cc(origin: dict, destination: dict) -> dict:
    """Copies the audience from the origin object to the destination object"""

    for key in ["to", "cc", "bto", "bcc", "audience"]:
        if key in origin:
            destination[key] = list(
                property_for_key_as_set(origin, key)
                | property_for_key_as_set(destination, key)
            )

    return destination


def as_list(value: dict | list | str) -> list:
    """Converts a value to a list by enclosing it

    ```pycon
    >>> as_list("http://example.com")
    ['http://example.com']

    >>> as_list(["http://example.com"])
    ['http://example.com']

    ```

    """

    if isinstance(value, list):
        return value

    return [value]


def combine_as_list(*values: list) -> list:
    """Combines the values into a list. Lists are combined,
    non list items are added as members

    ```pycon
    >>> combine_as_list("http://one.test", ["http://two.test"])
    ['http://one.test', 'http://two.test']

    ```
    """

    return sum((as_list(x) for x in values), [])



from typing import Any
from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field


class JrdLink(BaseModel):
    """
    See [RFC 7033](https://www.rfc-editor.org/rfc/rfc7033.html).

    ```pycon
    >>> JrdLink(rel="self", href="http://test.example/actor").model_dump(exclude_none=True)
    {'rel': 'self', 'href': 'http://test.example/actor'}

    ```
    """

    model_config = ConfigDict(
        extra="allow",
    )
    rel: str | None = Field(default=None, examples=["self"], description="rel")
    type: str | None = Field(default=None, examples=["application/activity+json"])
    href: str | None = Field(
        default=None,
        examples=["http://test.example/actor"],
        description="""
    when used with the 'href' attribute, conveys a link relation between the host described by the document and a common target URI
    """,
    )
    titles: dict[str, Any] | None = Field(
        default=None,
        description="titles",
    )
    properties: dict[str, Any] | None = Field(
        default=None,
        description="properties",
    )
    template: str | None = Field(
        default=None,
        description="template attribute conveys a relation whose context is an individual resource within the host-meta document scope,",
    )


class JrdData(BaseModel):
    """
    See [RFC 6415](https://www.packetizer.com/rfc/rfc6415/)

    ```pycon
    >>> JrdData(subject="acct:actor@test.example").model_dump(exclude_none=True)
    {'subject': 'acct:actor@test.example'}

    ```
    """

    model_config = ConfigDict(
        extra="allow",
    )
    subject: str | None = Field(None, examples=["acct:actor@test.example"])
    expires: datetime | None = Field(
        default=None,
        description="expiration date time",
    )
    aliases: list[str] | None = Field(
        default=None,
        description="value a string array containing the values of each element in order",
    )
    properties: dict[str, Any] | None = Field(
        default=None,
        description="value an object with each element included as a name/value pair with the value of the type attribute as name, and element value included as a string value.",
    )
    links: list[JrdLink] | None = Field(
        default=None,
        description="a single name/value pair with the name 'links', and value an array with each element included as an object",
    )



# generated by datamodel-codegen:
#   filename:  2.0
#   timestamp: 2024-12-14T18:04:04+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List

from pydantic import BaseModel, ConfigDict, Field, conint, constr


class Software(BaseModel):
    """
    Metadata about server software in use.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    name: constr(pattern=r"^[a-z0-9-]+$") = Field(
        description="""
    The canonical name of this server software.
    """
    )
    version: str = Field(
        description="""
    The version of this server software.
    """
    )


class Protocol(str, Enum):
    activitypub = "activitypub"
    buddycloud = "buddycloud"
    dfrn = "dfrn"
    diaspora = "diaspora"
    libertree = "libertree"
    ostatus = "ostatus"
    pumpio = "pumpio"
    tent = "tent"
    xmpp = "xmpp"
    zot = "zot"


class InboundEnum(str, Enum):
    atom1_0 = "atom1.0"
    gnusocial = "gnusocial"
    imap = "imap"
    pnut = "pnut"
    pop3 = "pop3"
    pumpio = "pumpio"
    rss2_0 = "rss2.0"
    twitter = "twitter"


class OutboundEnum(str, Enum):
    atom1_0 = "atom1.0"
    blogger = "blogger"
    buddycloud = "buddycloud"
    diaspora = "diaspora"
    dreamwidth = "dreamwidth"
    drupal = "drupal"
    facebook = "facebook"
    friendica = "friendica"
    gnusocial = "gnusocial"
    google = "google"
    insanejournal = "insanejournal"
    libertree = "libertree"
    linkedin = "linkedin"
    livejournal = "livejournal"
    mediagoblin = "mediagoblin"
    myspace = "myspace"
    pinterest = "pinterest"
    pnut = "pnut"
    posterous = "posterous"
    pumpio = "pumpio"
    redmatrix = "redmatrix"
    rss2_0 = "rss2.0"
    smtp = "smtp"
    tent = "tent"
    tumblr = "tumblr"
    twitter = "twitter"
    wordpress = "wordpress"
    xmpp = "xmpp"


class Services(BaseModel):
    """
    The third party sites this server can connect to via their application API.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    inbound: List[InboundEnum] = Field(
        min_length=0,
        description="""
    The third party sites this server can retrieve messages from for combined display with regular traffic.
    """,
    )
    outbound: List[OutboundEnum] = Field(
        min_length=0,
        description="""
    The third party sites this server can publish messages to on the behalf of a user.
    """,
    )


class Users(BaseModel):
    """
    statistics about the users of this server.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    total: conint(ge=0) = Field(
        0,
        description="""
    The total amount of on this server registered users.
    """,
    )
    activeHalfyear: conint(ge=0) = Field(
        0,
        description="""
    The amount of users that signed in at least once in the last 180 days.
    """,
    )
    activeMonth: conint(ge=0) = Field(
        0,
        description="""
    The amount of users that signed in at least once in the last 30 days.
    """,
    )


class Usage(BaseModel):
    """
    Usage statistics for this server.
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    users: Users = Field(
        default_factory=Users,
        description="""
    statistics about the users of this server.
    """,
    )
    localPosts: conint(ge=0) = Field(
        0,
        description="""
    The amount of posts that were made by users that are registered on this server.
    """,
    )
    localComments: conint(ge=0) = Field(
        0,
        description="""
    The amount of comments that were made by users that are registered on this server.
    """,
    )


class NodeInfo(BaseModel):
    """
    NodeInfo schema version 2.0.

    Usage:
    ```pycon
    >>> NodeInfo(software={"name": "bovine", "version": "0.0.1"}).model_dump()
    {'version': '2.0',
        'software': {'name': 'bovine', 'version': '0.0.1'},
        'protocols': [<Protocol.activitypub: 'activitypub'>],
        'services': {'inbound': [],
            'outbound': []},
        'openRegistrations': False,
        'usage': {'users':
            {'total': 0, 'activeHalfyear': 0, 'activeMonth': 0},
            'localPosts': 0,
            'localComments': 0},
        'metadata': {}}

    ```
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    version: str = Field(
        "2.0",
        description="""
    The schema version, must be 2.0.
    """,
    )
    software: Software = Field(
        description="""
    Metadata about server software in use.
    """
    )
    protocols: List[Protocol] = Field(
        [Protocol.activitypub],
        min_length=1,
        description="""
    The protocols supported on this server.
    """,
    )
    services: Services = Field(
        Services(inbound=[], outbound=[]),
        description="""
    The third party sites this server can connect to via their application API.
    """,
    )
    openRegistrations: bool = Field(
        False,
        description="""
    Whether this server allows open self-registration.
    """,
    )
    usage: Usage = Field(
        default_factory=Usage,
        description="""
    Usage statistics for this server.
    """,
    )
    metadata: Dict[str, Any] = Field(
        {},
        description="""
    Free form key value pairs for software specific values. Clients should not rely on any specific key present.
    """,
    )



"""
This package contains a simple ActivityPub server for
testing purposes. See [Bovine - Tutorial-Server][https://bovine.readthedocs.io/en/latest/tutorials/server/].

"""

import os
import secrets
import json
import tomli_w
import tomllib

from dataclasses import dataclass, asdict

from quart import Quart, request, send_file

from bovine import BovineActor
from bovine.activitystreams import Actor
from bovine.crypto import build_validate_http_signature, generate_rsa_public_private_key
from bovine.crypto.types import CryptographicIdentifier
from bovine.utils import webfinger_response_json

from .config import public_key as config_public_key, private_key as config_private_key


@dataclass
class ServerConfig:
    protocol: str = "http"
    hostname: str = "bovine"
    handle_name: str = "buttercup"
    public_key: str = config_public_key
    private_key: str = config_private_key

    @property
    def actor_id(self) -> str:
        """Returns the uri of the actor:

        ```pycon
        >>> ServerConfig().actor_id
        'http://bovine/actor'

        ```
        """
        return f"{self.protocol}://{self.hostname}/actor"

    @property
    def icon(self) -> str:
        """Returns the uri of the icon:

        ```pycon
        >>> ServerConfig().icon
        'http://bovine/cow'

        ```
        """
        return f"{self.protocol}://{self.hostname}/cow"

    @property
    def webfinger_response(self):
        """The expected response for a webfinger lookup

        ```pycon
        >>> ServerConfig().webfinger_response
        {'subject': 'acct:buttercup@bovine',
            'links': [{'rel': 'self',
                'type': 'application/activity+json',
                'href': 'http://bovine/actor'}]}

        ```
        """

        return webfinger_response_json(
            f"acct:{self.handle_name}@{self.hostname}", self.actor_id
        )

    def make_id(self):
        return f"{self.protocol}://{self.hostname}/" + secrets.token_urlsafe(6)

    def create_actor_object(self) -> dict:
        return Actor(
            id=self.actor_id,
            preferred_username=self.handle_name,
            name="Buttercup the exemplary bovine",
            inbox=f"{self.protocol}://{self.hostname}/inbox",
            outbox=self.actor_id,
            public_key=self.public_key,
            public_key_name="main-key",
            icon={"mediaType": "image/png", "type": "Image", "url": self.icon},
        ).build()

    def create_actor(self) -> BovineActor:
        """Returns the BovineActor"""
        return BovineActor(
            actor_id=self.actor_id,
            public_key_url=f"{self.actor_id}#main-key",
            secret=self.private_key,
        )

    def save(self, filename):
        with open(filename, "wb") as fp:
            tomli_w.dump(asdict(self), fp, multiline_strings=True)

    @staticmethod
    def load(filename):
        with open(filename, "rb") as fp:
            data = tomllib.load(fp)
        return ServerConfig(**data)

    @staticmethod
    def from_env():
        protocol = os.environ.get("BOVINE_TEST_PROTOCOL", "http")
        hostname = os.environ.get("BOVINE_TEST_HOSTNAME", "bovine")
        handle_name = os.environ.get("BOVINE_TEST_NAME", "buttercup")

        public_key, private_key = generate_rsa_public_private_key()

        return ServerConfig(
            protocol=protocol,
            hostname=hostname,
            handle_name=handle_name,
            public_key=public_key,
            private_key=private_key,
        )


def create_app(config: ServerConfig) -> Quart:
    app = Quart(__name__)
    actor = config.create_actor()
    actor_object = config.create_actor_object()

    @app.before_serving
    async def startup():
        await actor.init()

    async def fetch_public_key(url):
        result = await actor.get(url)
        return CryptographicIdentifier.from_public_key(result["publicKey"])

    verify = build_validate_http_signature(fetch_public_key)

    @app.get("/.well-known/webfinger")
    async def webfinger():
        return config.webfinger_response

    @app.get("/actor")
    async def get_actor():
        return actor_object, 200, {"content-type": "application/activity+json"}

    @app.post("/inbox")
    async def post_inbox():
        controller = await verify(request)
        if not controller:
            return "ERROR", 401

        data = await request.get_json()
        print(f"Received in inbox from {controller}")
        print(json.dumps(data, indent=2))
        print()
        return "success", 202

    @app.get("/cow")
    async def cow():
        path = "/".join(__file__.split("/")[:-1] + ["resources", "logo.png"])
        return await send_file(path)

    return app



import jsonschema
import json

from behave import then
from behave.api.async_step import async_run_until_complete


@then('The result obeys the schema given by "{schema_url}"')
@async_run_until_complete
async def validate_against_schema(context, schema_url):
    """
    Checks that the result stored in `context.result`
    satisfies the schema retrieved from `schema_url`.

    ```gherkin
    Then The result obeys the schema given by "https://domain.example/schema.json"
    ```
    """
    response = await context.session.get(
        schema_url, headers={"accept": "application/json"}
    )
    schema = json.loads(await response.text())

    jsonschema.validate(context.result, schema)


@then('The result has the type "{activity_type}"')
def check_type(context, activity_type):
    """
    Checks that the result stored in `context.result`
    has the `type` property given by `activity_type`.

    ```gherkin
    Then The result has the type "Follow"
    ```
    """
    assert context.result.get("type") == activity_type



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

"""The goal of the testing module is to provide
some examples of commonly used things to make writing
tests easier.


<!--
```
>>> import bovine.utils
>>> bovine.utils.now_isoformat = lambda: '2024-09-26T18:35:42Z'

```
-->

## Usage examples

For [public_key][bovine.testing.public_key] and [private_key][bovine.testing.public_key]

```pycon
>>> from bovine.crypto.helper import pem_private_key_to_pem_public_key
>>> pem_private_key_to_pem_public_key(private_key) == public_key
True

```

For [ed25519_key][bovine.testing.ed25519_key] and
[did_key][bovine.testing.did_key]

```pycon
>>> from bovine.crypto import private_key_to_did_key
>>> private_key_to_did_key(ed25519_key) == did_key
True

```


For [activity_factory][bovine.testing.activity_factory] see also
[ActivityFactory][bovine.activitystreams.activity_factory.ActivityFactory]

```pycon
>>> activity_factory.like("http://bovine.example").build()
{'@context': 'https://www.w3.org/ns/activitystreams',
    'type': 'Like',
    'actor': 'http://actor.example',
    'published': '2024-09-26T18:35:42Z',
    'object': 'http://bovine.example'}

```

For [object_factory][bovine.testing.object_factory] see also
[ObjectFactory][bovine.activitystreams.object_factory.ObjectFactory]
and the provided [note][bovine.testing.note]

```pycon
>>> object_factory.note(content="moo").build()
{'@context': 'https://www.w3.org/ns/activitystreams',
    'type': 'Note',
    'attributedTo': 'http://actor.example',
    'published': '2024-09-26T18:35:42Z',
    'content': 'moo'}

>>> note
{'@context': 'https://www.w3.org/ns/activitystreams',
    'type': 'Note',
    'attributedTo': 'http://actor.example',
    'to': ['https://www.w3.org/ns/activitystreams#Public'],
    'cc': ['http://actor.example/followers'],
    'published': '...',
    'content': 'moo'}


```

"""

import secrets

from bovine.activitystreams.activity_factory import ActivityFactory
from bovine.activitystreams.object_factory import ObjectFactory

from .config import public_key as public_key_const, private_key as private_key_const

public_key = public_key_const
"""The public key corresponding to [private_key][bovine.testing.private_key]
"""

private_key = private_key_const
"""An RSA private key"""

ed25519_key = "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
"""An Ed25519 private key encoded in multibase"""

public_key_multibase = "z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"
"""A public ed25519 key encoded in multibase"""

did_key = f"did:key:{public_key_multibase}"
"""A did key"""

ecp256_key = "z42twTcNeSYcnqg1FLuSFs2bsGH3ZqbRHFmvS9XMsYhjxvHN"
"""A NIST P-256 private key"""

actor = {
    "@context": [
        "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security/v1",
    ],
    "publicKey": {
        "id": "http://actor.example#name",
        "owner": "http://actor.example",
        "publicKeyPem": public_key,
    },
    "id": "http://actor.example",
    "type": "Person",
    "inbox": "http://actor.example/inbox",
    "outbox": "http://actor.example/outbox",
    "followers": "http://actor.example/followers",
}
"""Example of an ActivityStreams Actor"""


def make_id():
    return "http://actor.example/" + secrets.token_urlsafe(8)


activity_factory = ActivityFactory(actor)
"""The activity factory for the actor"""

activity_factory_id = ActivityFactory(actor, id_generator=make_id)
"""The activity factory for the actor with predefined id generator"""


object_factory = ObjectFactory(actor)
"""The object factory for the actor"""

note = object_factory.note(content="moo").as_public().build()
"""Example of a Note object"""



"""
This module provides routines to create PropertyValue objects
as used by [Mastodon](https://docs.joinmastodon.org/spec/activitypub/#schema).
We try to implement json-ld as we understand it.

We use PropertyValue as defined by schema.org [here](https://schema.org/PropertyValue).
"""

property_value_context = {
    "PropertyValue": {
        "@id": "https://schema.org/PropertyValue",
        "@context": {
            "value": "https://schema.org/value",
            "name": "https://schema.org/name",
        },
    }
}
"""If using PropertyValue object, include this in `@context`"""


def from_key_value(key, value):
    """Creates a PropertyValue object from a key and a value

    ```pycon
    >>> from_key_value("name", "Alice")
    {'type': 'PropertyValue', 'name': 'name', 'value': 'Alice'}

    ```
    """

    return {
        "type": "PropertyValue",
        "name": key,
        "value": value,
    }


def from_dictionary(values: dict) -> list[dict]:
    """Creates a list of PropertyValue objects from a dictionary

    ```pycon
    >>> from_dictionary({"name": "Alice", "job": "Sending secret messages"})
    [{'type': 'PropertyValue', 'name': 'name', 'value': 'Alice'},
    {'type': 'PropertyValue', 'name': 'job', 'value': 'Sending secret messages'}]

    ```
    """

    return [from_key_value(key, value) for key, value in values.items()]


def dictionary_to_actor_attachment(values: dict) -> dict:
    """Creates a stub actor with the property values corresponding
    to `values` as attachment

    ```pycon
    >>> dictionary_to_actor_attachment({"name": "Alice"})
    {'@context': ['https://www.w3.org/ns/activitystreams',
        {'PropertyValue': {'@id': 'https://schema.org/PropertyValue',
            '@context': {'value': 'https://schema.org/value',
                'name': 'https://schema.org/name'}}}],
        'attachment': [{'type': 'PropertyValue',
            'name': 'name',
            'value': 'Alice'}]}

    ```

    """
    return {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            property_value_context,
        ],
        "attachment": from_dictionary(values),
    }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import json
import warnings


from urllib.parse import urlparse
from dataclasses import dataclass, field
from typing import Callable, Awaitable, Tuple
import bovine.utils
from bovine.utils import parse_gmt

from .http_signature import HttpSignature
from .signature import Signature, RFC9421Signature
from .types import CryptographicIdentifier

from .digest import validate_digest

logger = logging.getLogger(__name__)


@dataclass
class SignatureChecker:
    """Dataclass to encapsulate the logic of checking a HTTP signature"""

    key_retriever: Callable[
        [str], Awaitable[Tuple[str | None, str | None] | CryptographicIdentifier | None]
    ] = field(
        metadata={
            "description": "used to resolve the keyId to the cryptographic information"
        }
    )
    skip_digest_check: bool = field(
        default=False, metadata={"description": "Set to true to skip digest check"}
    )

    async def handle_rfc9421_signature(
        self,
        method: str,
        url: str,
        headers: dict,
        body: Callable[[], Awaitable[str | bytes]],
    ) -> str | None:
        parsed_signature = RFC9421Signature.from_headers(
            headers["signature-input"], headers["signature"]
        )
        key_result = await self.key_retriever(parsed_signature.key_id)

        if key_result is None:
            logger.info(f"Could not retrieve key from {parsed_signature.key_id}")
            return None

        http_signature = self.build_rfc9421_http_signature(
            parsed_signature, method, url, headers
        )
        return http_signature.verify_with_identity_quoted(
            key_result, parsed_signature.signature
        )

    async def validate_signature(
        self,
        method: str,
        url: str,
        headers: dict,
        body: Callable[[], Awaitable[str | bytes]] | None,
    ) -> str | None:
        """Valids a given signature

        :param method: The http method either get or post
        :param url: The url being queried
        :param headers: The request headers
        :param body: A coroutine resolving the the request body. Used for post requests to check the digest.

        :returns:
        """
        if "signature" not in headers:
            logger.debug("Signature not present on request for %s", url)
            logger.debug(json.dumps(dict(headers)))
            return None

        if method.lower() == "post" and not self.skip_digest_check:
            if not self.validate_digest(headers, await body()):
                logger.warning("Validating digest failed")
                return None

        if "signature-input" in headers:
            logger.info("RFC 9421 Signature")
            return await self.handle_rfc9421_signature(method, url, headers, body)

        try:
            parsed_signature = Signature.from_signature_header(headers["signature"])
            signature_fields = parsed_signature.fields

            if (
                "(request-target)" not in signature_fields
                or "date" not in signature_fields
            ):
                logger.warning("Required field not present in signature")
                return None

            if method.lower() == "post" and all(
                field not in signature_fields for field in ["digest", "content-digest"]
            ):
                logger.warning("Digest not present, but computable")
                return None

            http_date = parse_gmt(headers["date"])
            if not bovine.utils.check_max_offset_now(http_date):
                logger.warning(f"Encountered invalid http date {headers['date']}")
                return None

            key_result = await self.key_retriever(parsed_signature.key_id)
            if isinstance(key_result, tuple):
                warnings.warn(
                    "Returning a tuple from key_retriever is deprecated, return a CryptographicIdentifier instead, will be remove in bovine 0.6.0",
                    DeprecationWarning,
                )
                logger.warn(
                    "Returning a tuple from key_retriever is deprecated, return a CryptographicIdentifier instead, will be remove in bovine 0.6.0"
                )
                key_result = CryptographicIdentifier.from_pem(*key_result)

            if key_result is None:
                logger.debug(f"Could not retrieve key from {parsed_signature.key_id}")
                return None

            http_signature = self.build_http_signature(
                signature_fields, method, url, parsed_signature, headers
            )

            return http_signature.verify_with_identity(
                key_result, parsed_signature.signature
            )

        except Exception as e:
            logger.exception(str(e))
            logger.error(headers)
            return None

        return None

    def build_http_signature(
        self, signature_fields, method, url, parsed_signature, headers
    ):
        http_signature = HttpSignature()
        for parsed_field in signature_fields:
            match parsed_field:
                case "(request-target)":
                    method = method.lower()
                    parsed_url = urlparse(url)
                    path = parsed_url.path
                    http_signature.with_field(parsed_field, f"{method} {path}")
                case "(expires)":
                    http_signature.with_field(parsed_field, parsed_signature.expires)
                case "(created)":
                    http_signature.with_field(parsed_field, parsed_signature.created)
                case _:
                    http_signature.with_field(parsed_field, headers[parsed_field])
        return http_signature

    def build_rfc9421_http_signature(self, parsed_signature, method, url, headers):
        http_signature = HttpSignature()
        for parsed_field in parsed_signature.fields:
            match parsed_field:
                case "@method":
                    http_signature.with_field(parsed_field, method.upper())
                case "@path":
                    parsed_url = urlparse(url)
                    path = parsed_url.path
                    http_signature.with_field(parsed_field, path)
                case "@authority":
                    http_signature.with_field(parsed_field, headers["host"])
                case _:
                    try:
                        http_signature.with_field(parsed_field, headers[parsed_field])
                    except Exception as error:
                        logger.exception(error)

        http_signature.with_field(
            "@signature-params", parsed_signature.signature_params
        )

        return http_signature

    async def validate_signature_request(self, request) -> str | None:
        """Validates a given signature

        :param request: The request object"""
        return await self.validate_signature(
            request.method, request.url, request.headers, request.get_data
        )

    def validate_digest(self, headers: dict, body: bytes) -> bool:
        if isinstance(body, str):
            warnings.warn("Got body of type str expected bytes")
            body = body.encode("utf-8")

        return validate_digest(headers, body)



"""
This optional package contains methods to create
objects used for media attachments. To use these
methods please run

```bash
pip install bovine[media]
```

As work on a media attachment is in progress, the
methods in this package might still change.
"""

try:
    from PIL import Image
    import cv2
    import audioread
    import isodate
except Exception:
    print("Please install media dependencies via pip install bovine[media]")

from dataclasses import dataclass
from datetime import timedelta
from typing import List
import base64

from bovine.activitystreams.object_factory import Object
from bovine.crypto.digest import digest_multibase


def file_props(filename: str, return_bytes_as_base64: bool = False):
    """
    Properties of a file

    ```pycon
    >>> file_props("./resources/test/cow.jpg")
    ('zQmVm2NjMCsrzFWFRbhQJwHNzZH8gesampoGpZcsSii12VL', 15356)

    ```
    """
    with open(filename, "rb") as fp:
        image_bytes = fp.read()

    file_size = len(image_bytes)
    digest = digest_multibase(image_bytes)

    if return_bytes_as_base64:
        return digest, file_size, base64.b64encode(image_bytes).decode()

    return digest, file_size


def iso8601_duration(seconds):
    """
    Converts seconds to iso 8601 duration

    ```pycon
    >>> iso8601_duration(3789)
    'PT1H3M9S'

    ```
    """

    return isodate.duration_isoformat(timedelta(seconds=seconds))


def image_to_object(image_uri, image, alt_text, digest, file_size):
    return Object(
        type="Image",
        url=image_uri,
        height=image.height,
        width=image.width,
        media_type=image.get_format_mimetype(),
        name=alt_text,
        digest_multibase=digest,
        file_size=file_size,
    )


def create_image(image_source: str, image_uri: str, alt_text: str | None = None):
    """
    Creates an Image object for the file

    ```pycon
    >>> create_image(
    ...     "./resources/test/cow.jpg",
    ...     "http://local.test/image.jpg",
    ...     alt_text="A beautiful cow").build()
    {'@context': ['https://www.w3.org/ns/activitystreams',
            'https://www.w3.org/ns/credentials/v2',
            {'size': 'https://joinpeertube.org/ns#size'}],
        'type': 'Image',
        'name': 'A beautiful cow',
        'url': 'http://local.test/image.jpg',
        'width': 200,
        'height': 164,
        'mediaType': 'image/jpeg',
        'digestMultibase': 'zQmVm2NjMCsrzFWFRbhQJwHNzZH8gesampoGpZcsSii12VL',
        'size': 15356}

    ```
    """

    digest, file_size = file_props(image_source)
    image = Image.open(image_source)

    return image_to_object(image_uri, image, alt_text, digest, file_size)


def create_data_uri(content_type: str, content_base64: str):
    """Creates a data uri

    ```pycon
    >>> create_data_uri("image/jpeg", "ABC")
    'data:image/jpeg;base64,ABC'

    ```
    """

    return f"data:{content_type};base64,{content_base64}"


def create_image_inline(image_source: str, alt_text: str | None = None):
    """
    Creates an Image object for the file

    ```pycon
    >>> create_image_inline(
    ...     "./resources/test/cow.jpg",
    ...     alt_text="A beautiful cow").build()
    {'@context': ['https://www.w3.org/ns/activitystreams',
            'https://www.w3.org/ns/credentials/v2',
            {'size': 'https://joinpeertube.org/ns#size'}],
        'type': 'Image',
        'name': 'A beautiful cow',
        'url': 'data:image/jpeg...
        'width': 200,
        'height': 164,
        'mediaType': 'image/jpeg',
        'digestMultibase': 'zQmVm2NjMCsrzFWFRbhQJwHNzZH8gesampoGpZcsSii12VL',
        'size': 15356}

    ```
    """

    digest, file_size, image_base64 = file_props(
        image_source, return_bytes_as_base64=True
    )
    image = Image.open(image_source)

    return image_to_object(
        create_data_uri(image.get_format_mimetype(), image_base64),
        image,
        alt_text,
        digest,
        file_size,
    )


@dataclass
class VideoInformation:
    digest: str
    file_size: int

    duration: str
    width: int
    height: int

    @staticmethod
    def from_video_source(video_source):
        digest, file_size = file_props(video_source)

        video = cv2.VideoCapture(video_source)
        fps = video.get(cv2.CAP_PROP_FPS)
        frame_count = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
        duration = frame_count / fps

        width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))

        return VideoInformation(
            digest=digest,
            file_size=file_size,
            duration=iso8601_duration(duration),
            width=width,
            height=height,
        )


def make_link(info, video_uri):
    return {
        "type": "Link",
        "size": info.file_size,
        "digest": info.digest,
        "width": info.width,
        "height": info.height,
        "href": video_uri,
        "mediaType": "video/mp4",
    }


def create_video(
    video_source: str | List[str], video_uri: str, alt_text: str | None = None
):
    """
    Creates a video object
    ```
    >>> create_video(
    ...     "./resources/test/cow_eating.mp4",
    ...     "http://local.test/video.mp4",
    ...     alt_text="A beautiful cow eating").build()
    {'@context':
        ['https://www.w3.org/ns/activitystreams',
        'https://www.w3.org/ns/credentials/v2',
        {'size': 'https://joinpeertube.org/ns#size'}],
        'type': 'Video',
        'name': 'A beautiful cow eating',
        'url': 'http://local.test/video.mp4',
        'width': 256,
        'height': 144,
        'mediaType': 'video/mp4',
        'digestMultibase': 'zQmSzK5qEe5tpjwGMhmjx9RvVoPkWhEmCwxP2s7wPMpKMoK',
        'size': 54373,
        'duration': 'PT3S'}

    ```

    This method also supports two video sources:

    ```
    >>> create_video(
    ...     ["./resources/test/cow_eating.mp4", "./resources/test/cow_eating_hd.mp4"],
    ...     ["http://local.test/video.mp4", "http://local.test/video_hd.mp4"],
    ...     alt_text="A beautiful cow eating").build()
    {'@context': ['https://www.w3.org/ns/activitystreams',
            'https://www.w3.org/ns/credentials/v2',
            {'size': 'https://joinpeertube.org/ns#size'}],
        'type': 'Video',
        'name': 'A beautiful cow eating',
        'url': [{'type': 'Link',
            'size': 54373,
            'digest': 'zQmSzK5qEe5tpjwGMhmjx9RvVoPkWhEmCwxP2s7wPMpKMoK',
            'width': 256,
            'height': 144,
            'href': 'http://local.test/video.mp4',
            'mediaType': 'video/mp4'},
        {'type': 'Link',
            'size': 2271723,
            'digest': 'zQme2X4rgWuRdmAtGGMSEbdoeRQ2NAL2VptcdRGTYDZbSKG',
            'width': 1920,
            'height': 1080,
            'href': 'http://local.test/video_hd.mp4',
            'mediaType': 'video/mp4'}],
        'duration': 'PT3S'}

    ```
    """

    if isinstance(video_source, list):
        video_infos = [VideoInformation.from_video_source(vs) for vs in video_source]
        links = [make_link(vi, vu) for vi, vu in zip(video_infos, video_uri)]
        digest, file_size, width, height, media_type = [None] * 5
        duration = video_infos[0].duration
    else:
        links = video_uri
        info = VideoInformation.from_video_source(video_source)
        digest, file_size = info.digest, info.file_size
        width, height = info.width, info.height
        duration = info.duration
        media_type = "video/mp4"

    return Object(
        type="Video",
        url=links,
        width=width,
        height=height,
        name=alt_text,
        digest_multibase=digest,
        file_size=file_size,
        duration=duration,
        media_type=media_type,
    )


def create_audio(audio_source: str, audio_uri: str, alt_text: str | None = None):
    """
    Creates an audio object
    ```pycon
    >>> create_audio("./resources/test/cow_moo.mp3",
    ...     "http://local.test/audio.mp3",
    ...     alt_text="A cow mooing").build()
    {'@context': ['https://www.w3.org/ns/activitystreams',
            'https://www.w3.org/ns/credentials/v2',
            {'size': 'https://joinpeertube.org/ns#size'}],
        'type': 'Audio',
        'name': 'A cow mooing',
        'url': 'http://local.test/audio.mp3',
        'mediaType': 'audio/mpeg',
        'digestMultibase': 'zQmSXTyLCPqoiGoUUwKRMKgFdddaAUkvQNr29nhB6tahb9Z',
        'size': 67709,
        'duration': 'PT2.1S'}

    ```
    """

    digest, file_size = file_props(audio_source)
    with audioread.audio_open(audio_source) as f:
        duration = f.duration
    # FIXME Wrong duration format

    return Object(
        type="Audio",
        url=audio_uri,
        name=alt_text,
        digest_multibase=digest,
        file_size=file_size,
        duration=iso8601_duration(duration),
        media_type="audio/mpeg",
    )



import pytest

import json
import jsonschema

from .media import create_image, create_video, create_audio, create_image_inline


def should_skip():
    try:
        import cv2  # noqa

        return False
    except Exception:
        print("EXCEPTION SKIPPING TESTS")
        return True


@pytest.mark.skipif(should_skip(), reason="Needs cv2 installed")
def test_image_load():
    uri = "http://local.test/image.jpg"
    result = create_image(
        "./resources/test/cow.jpg", uri, alt_text="A beautiful cow"
    ).build()

    assert result["type"] == "Image"
    assert result["url"] == uri
    assert result["width"] == 200
    assert result["mediaType"] == "image/jpeg"
    assert result["name"] == "A beautiful cow"

    with open("./resources/media.schema.json") as fp:
        schema = json.load(fp)

    jsonschema.validate(result, schema)


@pytest.mark.skipif(should_skip(), reason="Needs cv2 installed")
def test_create_image_inline():
    result = create_image_inline(
        "./resources/test/cow.jpg", alt_text="A beautiful cow"
    ).build()

    assert result["type"] == "Image"
    assert result["url"].startswith("data:image/jpeg;base64,")
    assert result["width"] == 200
    assert result["mediaType"] == "image/jpeg"
    assert result["name"] == "A beautiful cow"

    with open("./resources/media.schema.json") as fp:
        schema = json.load(fp)

    jsonschema.validate(result, schema)


@pytest.mark.skipif(should_skip(), reason="Needs cv2 installed")
def test_create_video():
    uri = "http://local.test/video.mp4"
    result = create_video(
        "./resources/test/cow_eating.mp4", uri, alt_text="A beautiful cow eating"
    ).build()

    assert result["type"] == "Video"
    assert result["url"] == uri
    assert result["width"] == 256
    assert result["mediaType"] == "video/mp4"
    assert result["name"] == "A beautiful cow eating"

    with open("./resources/media.schema.json") as fp:
        schema = json.load(fp)

    jsonschema.validate(result, schema)


@pytest.mark.skipif(should_skip(), reason="Needs cv2 installed")
def test_create_multi_video():
    result = create_video(
        ["./resources/test/cow_eating.mp4", "./resources/test/cow_eating_hd.mp4"],
        ["http://local.test/video.mp4", "http://local.test/video_hd.mp4"],
        alt_text="A beautiful cow eating",
    ).build()

    assert result["type"] == "Video"
    assert result["name"] == "A beautiful cow eating"

    assert isinstance(result["url"], list)

    with open("./resources/media.schema.json") as fp:
        schema = json.load(fp)

    jsonschema.validate(result, schema)


@pytest.mark.skipif(should_skip(), reason="Needs cv2 installed")
def test_create_audio():
    uri = "http://local.test/audio.mp3"
    result = create_audio(
        "./resources/test/cow_moo.mp3", uri, alt_text="A cow mooing"
    ).build()

    assert result["type"] == "Audio"
    assert result["url"] == uri
    assert result["mediaType"] == "audio/mpeg"
    assert result["name"] == "A cow mooing"

    with open("./resources/media.schema.json") as fp:
        schema = json.load(fp)

    jsonschema.validate(result, schema)



from quart import Quart

from .server import ServerConfig, create_app


def test_create_app():
    config = ServerConfig()

    app = create_app(config)

    assert isinstance(app, Quart)


def test_save_config(tmp_path):
    config = ServerConfig(handle_name="vanilla")

    config.save(tmp_path / "config.toml")

    new_config = ServerConfig.load(tmp_path / "config.toml")

    assert new_config.handle_name == "vanilla"


def test_actor_object():
    config = ServerConfig()

    actor_object = config.create_actor_object()

    assert actor_object["name"] == "Buttercup the exemplary bovine"



"""
Helpers for BDD. To use the helpers that provide
a session use an environment file
such as

```python title="features/environment.py"
--8<-- "./features/environment.py"
```

"""

import aiohttp

import asyncio


def before_all(context):
    """Creates necessary variables to store stuff in"""
    context.session = None
    context.responses = {}


async def create_client_session(context):
    context.session = aiohttp.ClientSession()


def before_scenario(context, scenario):
    """Ensures an [aiohttp.ClientSession][] is present"""
    if context.session is None:
        asyncio.get_event_loop().run_until_complete(create_client_session(context))

    context.responses = {}


def after_scenario(context, scenario):
    """Closes the [aiohttp.ClientSession][]  if present"""
    if context.session:
        asyncio.get_event_loop().run_until_complete(context.session.close())
        context.session = None
